/* This file is generated by "mtxrun --script "mtx-wtoc.lua" from the metapost cweb files. */


# ifndef MPMATHPOSIT_H
# define MPMATHPOSIT_H 1

# include "mp.h"
# include "softposit.h"

math_data *mp_initialize_posit_math (MP mp);

# endif

