/*tex

    For a while I keep some recently changed code here. Just in case ... we can always look in
    the GIT history if needed. Much more happened but that code has gone already. This is from
    end januari 2021. Eventually all here goes (or already went) into the void, including code
    that has been rewritten, pruned, upgraded but eventually got discarded.  The fittest code
    survives.

*/

/* pagebuilder state experiment in 2021-05-23 zip */