/* This file is generated by "mtxrun --script "mtx-wtoc.lua" from the metapost cweb files. */


# include <stdio.h>

# include "mpconfig.h"
# include "mpmathbinary.h"

extern void tex_normal_warning (const char *t, const char *p);

math_data *mp_initialize_binary_math (MP mp)
{
    (void) (mp);
    tex_normal_warning("mplib", "binary mode is not available.");
    return NULL;
}
