/* This file is generated by "mtxrun --script "mtx-wtoc.lua" from the metapost cweb files. */


# ifndef MPMATHDECIMAL_H
# define MPMATHDECIMAL_H 1

# include "mp.h"

math_data *mp_initialize_decimal_math (MP mp);

# endif

